/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include <stdint.h> // Standard Integer Types
#include <stdio.h> // Standard IO facilities
#include <stdlib.h> // General utilities
#include <stdbool.h> // Boolean
#include <string.h> // Strings

#include <avr/io.h> // AVR device-specific IO definitions
#include <util/delay.h> // Convenience functions for busy-wait delay loops
#include <avr/interrupt.h> // Interrupts
#include <avr/wdt.h> // Watchdog timer handling
#include <avr/pgmspace.h> // Program Space Utilities
#include <avr/sleep.h> // Power Management and Sleep Modes


#include "uart.h" // basic UART functions
#include "main.h" // main definitions

/* variables */
volatile uint8_t state_portD; // the state of port D
volatile bool scale_on = false; // is the scale on (based on the SCALE_ON PIN)
volatile bool pwm_high = false; // is the PWM for the last weight measurement high
volatile bool measurement_flag = false; // is a PWM  measurement value ready
volatile uint16_t measurement_value = 0; // the PWM measurement value from the timer

/* switch off LED */
void led_off(void)
{
	PORTB &= ~(1<<LED); // remove power to LED
}

/* switch on LED */
void led_on(void)
{
	PORTB |= (1<<LED); // provide power to LED
}

/* toggle LED */
void led_toggle(void)
{
	PINB |= (1<<LED);
}

/* scale on interrupt
 * PCI2 Interrupt Vector for PCINT[23:16]/PORTD
 */
ISR(PCINT2_vect)
{
	if ((state_portD&(1<<SCALE_ON))!=(PIND&(1<<SCALE_ON))) { // scale on state changed
		if (PIND&(1<<SCALE_ON)) {
			scale_on = true;
			led_on();
		} else {
			scale_on = false;
			led_off();
		}
	}
	state_portD = PIND; // save new state
}

/* PWM weight signal
 * analog comparator interrupt
 */
ISR(ANALOG_COMP_vect)
{
	if (scale_on && !measurement_flag) { // only save value if scale is on and previous value is read
		measurement_value = TCNT1; // save value
		TCNT1 = 0; // reset timer		
		if (ACSR&(1<<ACO)) { // PWM raising edge
			pwm_high = false;
			led_on();
		} else { // PWM falling edge
			pwm_high = true;
			led_off();
		}
		measurement_flag = true; // signal new value is ready
	}
}

/* disable watchdog when booting */
void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));
void wdt_init(void)
{
	MCUSR = 0;
	wdt_disable();

	return;
}

/* initialize GPIO */
void io_init(void)
{
	/* use UART as terminal */
	uart_init();
	stdout = &uart_output;
	stdin  = &uart_input;

	/* gpio */
	/* LED */
	DDRB |= (1<<LED); // LED is driven by pin (set as output)
	led_off();

	/* scale on signal */
	DDRD &= ~(1<<SCALE_ON); // SIGNAL_ON is input (should be per default)
	PCIFR &= ~(1<<PCIF2); // clear interrupt flag for SCALE_ON on PCINT[23:16]/PORTD
	PCICR |= (1<<PCIE2); // enable interrupt for PCINT[23:16]/PORTD
	PCMSK2 |= (1<<PCINT21); // enable interrupt for SCALE_ON
	state_portD = PIND; // save state to detect change

	/* PWM scale weight signal on analog comparator */
	DDRD &= ~((1<<SCALE_PWM)|(1<<REF_PWM)); // analog comparator is input (should be per default)
	ADCSRA &= ~(1<<ADEN); // switch off ADC
	ADCSRB &= ~(1<<ACME); // disable analog comparator multiplexer, use AIN1 as negative input (should be per default)
	ACSR &= ~(1<<ACD); // enable analog comparator (should be per default)
	ACSR &= ~(1<<ACBG); // use AIN0 as positiv input (should be per default)
	ACSR |= (1<<ACI); // clear analog comparator interrup flag
	ACSR &= ~((1<<ACIS1)|(1<<ACIS0)); // comparator interrupt on output toggle
	DIDR1 |= ((1<<AIN1D)|(1<<AIN0D)); // disable digital input buffer on AIN0 and AIN1
	ACSR |= (1<<ACIE); // enable analog comparator interrupt

	/* use timer 1 (with capture unit) to measure PWM */
	TCCR1A &= ~((1<<WGM11)|(1<<WGM10)); // mode 0: normal (should be per default)
	TCCR1B &= ~((1<<WGM13)|(1<<WGM12)); // mode 0: normal (should be per default)
	TCCR1B |= (1<<CS12)|(0<<CS11)|(0<<CS10); // set prescale to 256 for 1.048576s before overflow (PWM should be 3Hz)

	sei(); /* enable interrupts */
}

int main(void)
{
	io_init(); // initialize IOs

	printf(PSTR("welcome to the body scale weight reader\n")); // print welcome message
	while (true) { // endless loop for micro-controller
		/* display value if available */
		if (measurement_flag) {
			if (pwm_high) {
				printf("high: %u\n",measurement_value);
			} else {
				printf("low: %u\n",measurement_value);
			}
			measurement_flag = false;
		}
		/* go to sleep and wait for next interrupt */
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();
	}
	return 0;
}

